# dedrm.py 1.0
# QTFairUse6 helper
# m4p + decrypted audio => m4a
# Copyright 2006 Igor Sckochinsky <skochinsky@mail.ru>
# based on m4a.py from Mutagen project:

# Copyright 2006 Joe Wreschnig <piman@sacredchao.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# $Id$

import struct
import sys

from cStringIO import StringIO

class MP4Error(IOError): pass

_CONTAINERS = ["moov", "udta", "trak", "mdia", "meta", "ilst", "stsd", "stbl",
               "minf", "stsd", "dinf", "sinf", "SINF", "schi", "drms", "drmi", "mp4a",
               '\251alb', '\251ART', '\251day', '\251nam', '\251cmt','\251too', '\251wrt', '----']

_SKIP_SIZE = { "meta": 4, "stsd": 8, "drms":28, "mp4a":28, "drmi":78}

__all__ = ['MP4','MP4Error']

def uint32(data):
    if type(data) == str: return struct.unpack('>L', data[:4])[0]
    else: return struct.pack('>L', data)

def uint64(data):
    if type(data) == str: return struct.unpack('>Q', data[:8])[0]
    else: return struct.pack('>Q', data)

class Atom(object):
    """An individual atom.

    Attributes:
    children -- list child atoms (or None for non-container atoms)
    length -- length of this atom, including length and name
    name -- four byte name of the atom, as a str
    offset -- location in the constructor-given fileobj of this atom
    data - atom contents (for leaves)
    skip_data - header data situated before leaves for containers
    """

    children = None

    def __init__(self, fileobj):
        self.offset = fileobj.tell()
        self.length, self.name = struct.unpack(">I4s", fileobj.read(8))
        l = self.length-8
        if self.length == 1:
            #raise MP4Error("64 bit atom sizes are not supported")
            self.length = struct.unpack(">Q", fileobj.read(8))[0]
            l -= 8
        elif self.length < 8:
            return

        #print "tag: %s, length: %d" % (self.name, self.length)
        if self.name in _CONTAINERS:
            self.children = []
            self.skip_data = fileobj.read(_SKIP_SIZE.get(self.name, 0))
            while fileobj.tell() < self.offset + self.length:
                self.children.append(Atom(fileobj))
        else:
            self.data = fileobj.read(l)

    def get_len(self):
        """get total atom length (including children and header)"""
        if self.name in _CONTAINERS:
            l = len(self.skip_data)
            for c in self.children:
                l+=c.get_len()
        else:
            if hasattr(self,'customRender'):
                l=self.length
            else:
                l=len(self.data)
        l+=8
        if l>=0x100000000: #more than 32 bits
            l+=8
        return l

    def render(self, f=None):
        """Render atom data, including children and header"""
        l = self.get_len()
        try:
            header = struct.pack(">I4s", l, self.name)
        except OverflowError:
            header = struct.pack(">I4sQ", 1, self.name, l)
        if self.name in _CONTAINERS:
            if f:
                f.write(header)
                f.write(self.skip_data)
                for c in self.children:
                    c.render(f)
            else:
                res = header + self.skip_data
                for c in self.children:
                    res += c.render()
                return res
        else:
            if f:
                f.write(header)
                f.write(self.data)
                if hasattr(self,'customRender'):
                    self.customRender(f)
            else:
                return header + self.data

    def __getitem__(self, remaining):
        """Look up a child atom, potentially recursively.

        e.g. atom['udta', 'meta'] => <Atom name='meta' ...>
        """
        if isinstance(remaining, basestring):
            remaining = remaining.split(".")
        if not remaining:
            return self
        elif self.children is None:
            raise KeyError("%r is not a container" % self.name)

        name = remaining[0]
        num = 1
        if len(name)>4:
            num = int(name[4:])
            name = name[0:4]
        cnt = 1
        for child in self.children:
            if child.name == name:
                if cnt == num:
                    return child[remaining[1:]]
                else:
                    cnt += 1
        else:
            raise KeyError, "%r not found" % remaining[0]

    def __delitem__(self, remaining):
        """Delete a child atom, potentially recursively.

        e.g. atom['udta', 'meta'] => <Atom name='meta' ...>
        """
        if self.children is None:
            raise KeyError("%r is not a container" % self.name)

        if isinstance(remaining, basestring):
            remaining = remaining.split(".")

        #print "deleting '%s' from '%s'" % (".".join(remaining), self.name)

        name = remaining[0]
        num = 1
        if len(name)>4:
            num = int(name[4:])
            name = name[0:4]
        cnt = 1
        for i in xrange(len(self.children)):
            if self.children[i].name == name:
                if cnt == num:
                    if len(remaining)>1:
                        del self.children[i][remaining[1:]]
                    else:
                        del self.children[i]
                    return
                else:
                    cnt += 1
        else:
            raise KeyError, "%r not found" % remaining[0]

    def __repr__(self):
        klass = self.__class__.__name__
        if self.children is None:
            return "<%s name=%r length=%r offset=%r>" % (
                klass, self.name, self.length, self.offset)
        else:
            children = "\n".join([" " + line for child in self.children
                                  for line in repr(child).splitlines()])
            return "<%s name=%r length=%r offset=%r\n%s>" % (
                klass, self.name, self.length, self.offset, children)

class Atoms(object):
    """Root atoms in a given file.

    Attributes:
    atoms -- a list of top-level atoms as Atom objects

    """
    def __init__(self, fileobj):
        self.atoms = []
        fileobj.seek(0, 2)
        end = fileobj.tell()
        fileobj.seek(0)
        while fileobj.tell() < end:
            a = Atom(fileobj)
            if a.name=='\x00'*4:
                break
            self.atoms.append(a)

    def path(self, *names):
        """Look up and return the complete path of an atom.

        For example, atoms.path('moov', 'udta', 'meta') will return a
        list of three atoms, corresponding to the moov, udta, and meta
        atoms.
        """
        path = [self]
        for name in names:
            path.append(path[-1][name,])
        return path[1:]

    def __getitem__(self, names):
        """Look up a child atom.

        'names' may be a list of atoms (['moov', 'udta']) or a string
        specifying the complete path ('moov.udta').
        """
        if isinstance(names, basestring):
            names = names.split(".")
        name = names[0]
        num = 1
        if len(name)>4:
            num = int(name[4:])
            name = name[0:4]
        cnt = 1
        for child in self.atoms:
            if child.name == name:
                if cnt == num:
                    return child[names[1:]]
                else:
                    cnt += 1
        else:
            raise KeyError, "%s not found" % names[0]

    def __repr__(self):
        return "\n".join([repr(child) for child in self.atoms])

    def render(self, f=None):
        res = ""
        for a in self.atoms:
            if f:
                a.render(f)
            else:
                res += a.render()
        return res

class MP4:
    """An MPEG-4 file"""

    def __init__(self, filename):
        self.filename = filename
        fileobj = file(filename, "rb")
        try:
            try:
                self.atoms = Atoms(fileobj)
            except:
                raise MP4Error("Error parsing file; not an MP4?")
            #print self.atoms
            #self.info = M4AInfo(self.atoms, fileobj)
            #print self.info.pprint()
            snd_num = 0
            num = 1
            while True:
                try:
                    hdlr = self.atoms["moov.trak%d.mdia.hdlr"%num]
                except KeyError:
                    break
                #print "Track %d: type '%s'"%(num,hdlr.data[8:12])
                if hdlr.data[8:12]=='soun':
                    snd_num = num
                num += 1

            if snd_num==0:
                raise MP4Error("No audio track found")

            self.trackno = snd_num
            #print "Audio track: %d"%snd_num
            stbl = self.atoms["moov.trak%d.mdia.minf.stbl"%snd_num]
            try:
                drms = stbl['stsd.drms']
            except KeyError:
                raise MP4Error("File not encrypted")

        finally:
            fileobj.close()

    def get_audio_len(self):
        """get expected audio data length"""
        stbl = self.atoms["moov.trak%d.mdia.minf.stbl"%self.trackno]
        stsz_data = stbl['stsz'].data
        sample_size = uint32(stsz_data[4:8])
        sample_count = uint32(stsz_data[8:12])
        totalsize = 0
        if sample_size==0:
            for i in xrange(sample_count):
                totalsize += uint32(stsz_data[12+i*4:16+i*4])
        else:
            totalsize = sample_size*sample_count
        return totalsize


    def dedrm(self,audio):
        """replace audio stream with contents of 'audio'"""
        stbl = self.atoms["moov.trak%d.mdia.minf.stbl"%self.trackno]
        drms = stbl['stsd.drms']
        frma = drms['sinf.frma']
        #print "Original format: %s"%frma.data
        drms.name = frma.data
        #print drms

        #parse atoms related to chunks and samples since we'll need to update the offsets

        stsz_data = stbl['stsz'].data
        sample_size = uint32(stsz_data[4:8])
        sample_count = uint32(stsz_data[8:12])
        sample_sizes = []
        if sample_size==0:
            for i in xrange(sample_count):
                sample_sizes.append(uint32(stsz_data[12+i*4:16+i*4]))
        else:
            sample_sizes = [sample_size,]*sample_count

        #print "Sample count: %d"%sample_count

        stco = stbl['stco']
        chunk_count = uint32(stco.data[4:8])
        #print "Chunk count: %d"%chunk_count
        first_chunk_offset = uint32(stco.data[8:12])

        stsc_data = stbl['stsc'].data
        entry_count = uint32(stsc_data[4:8])
        stsc = []
        off = 8
        for i in xrange(entry_count):
            stsc.append(struct.unpack(">III",stsc_data[off:off+12]))
            off += 12

        #print "Sample to chunk table entries: %d"%len(stsc)

        smp = 0
        smp_off = 0
        chunk_sizes = []
        for i in xrange(len(stsc)):
            c1, spc, sdi = stsc[i]
            if i+1<len(stsc):
                c2 = stsc[i+1][0]
            else:
                c2 = chunk_count+1
            c = c1
            while c<c2:
                chunk_size = sum(sample_sizes[smp:smp+spc])
                chunk_sizes.append(chunk_size)
                #print "Chunk %d: %d samples (size=%d), index %d" % (c, spc, chunk_size, sdi)
                #print "Offset: %08X (%08X)" % (chunk_offsets[c-1]+first_chunk_offset, chunk_offsets[c-1])
                smp_off += chunk_size
                smp += spc
                c+=1

        if len(chunk_sizes)!=chunk_count:
            raise MP4Error("Error parsing 'stsc'!")

        total_size = sum(chunk_sizes)
        if (type(audio) is str) and (len(audio)!=total_size):
            raise MP4Error("Decrypted audio size doesn't match! (got %d, expected %d)"%(len(audio),total_size))

        #delete FairPlay-related atoms
        #needs to be done BEFORE calculating new offsets
        del drms['sinf']
        ilst = self.atoms["moov.udta.meta.ilst"]
        for x in ('akID', 'apID', 'atID', 'cnID', 'geID', 'plID', 'xeID', 'sfID'):
            try:
                del ilst[x]
            except:
                #ignore non-existing keys
                pass

        #find a valid mdat and calculate its offset in the new file
        off = 0
        old_off = 0
        for a in self.atoms.atoms:
            if a.name=='mdat' and a.length>8 and first_chunk_offset>a.offset:
                break
            else:
                off += a.get_len()
        else:
            raise MP4Error("No valid 'mdat' found!")

        #print "New mdat offset: %08X"%off
        start_off = first_chunk_offset-a.offset-8
        #print "Stream start: %08X (in mdat: %04X)" % (off + start_off, start_off)
        first_off = off+start_off+8
        new_stco = uint32(0)+uint32(chunk_count)
        off = first_off
        for i in xrange(chunk_count):
            sz = chunk_sizes[i]
            #print "Chunk %d: size %d, offset %08X(%08X)" % (i+1,sz,off,off-first_off)
            new_stco += uint32(off)
            off += sz

        #print "End offset: %08X(%08X)" % (off,off-first_off)
        #replace mdat contents with decrypted stream
        a.data = a.data[0:start_off]
        if type(audio) is str:
            a.data += audio
        else:
            a.customRender = audio
            a.length = total_size+start_off
        #set new chunk offsets table
        stco.data = new_stco

        #print self.atoms

    def save(self, filename):
        """Save (updated) atom tree to a new file"""
        fileobj = file(filename,"wb")
        self.atoms.render(fileobj)
        fileobj.close()

import os.path
if __name__=='__main__':
    print "dedrm.py 1.1 Copyright 2006 Igor Skochinsky"
    if len(sys.argv)<2:
        print "Usage: dedrm.py file.m4p dumped.aac"
        sys.exit(1)
    try:
        f1 = os.path.splitext(sys.argv[1])
        if f1[1]!='.m4p':
            print "Input file is not an m4p!"
            sys.exit(1)
        infile = f1[0]+".m4p"
        outfile = f1[0]+".m4a"
        print 'Processing "%s"' % infile
        m = MP4(infile)
        m.dedrm(file(sys.argv[2],"rb").read())
        m.save(outfile)
        print 'Saved "%s" successfully' % outfile
    except MP4Error, e:
        print "Error: %s"%e
        sys.exit(2)
